/*
 * Resource server functionality.
 *
 * Copyright (C) 2018, 2019, 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/types.h>

#include <pthread-l4.h>
#include <pthread.h>

#include "resource_server.h"



/* Convenience server methods. */

/* Bind to a named IPC gate capability. */

long ResourceServer::bind(const char *name)
{
  return ipc_server_bind(name, (l4_umword_t) _resource, &_config->server);
}

/* Start in the same thread with no deletion notifications or finalisation. */

long ResourceServer::start()
{
  resource_init_config(_config, _resource);
  _config->thread = pthread_l4_cap(pthread_self());
  return resource_start_config(_config, _resource);
}

/* Start a new thread with deletion notifications and finalisation. */

long ResourceServer::start_thread()
{
  pthread_t thread;
  pthread_attr_t attr;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  resource_init_config(_config, _resource);
  pthread_create(&thread, &attr, ipc_server_start_mainloop, _config);
  resource_set_config_threaded(_config, pthread_l4_cap(thread), 1);

  return resource_start_config(_config, _resource);
}



/* Initialise a server configuration for a resource. */

void resource_init_config(ipc_server_config_type *config, Resource *resource)
{
  config->handler_obj = resource->interface();
  config->finaliser_obj = resource;
  config->expected_items = resource->expected_items();
  config->handler = resource->handler();
}

/* Set a configuration to be threaded. */

void resource_set_config_threaded(ipc_server_config_type *config,
                                  l4_cap_idx_t thread, int new_thread)
{
  config->finaliser = resource_thread_finaliser;
  config->new_thread = new_thread;
  config->thread = thread;
  config->notifications = 1;
}

/* Activate a resource and start a server for it. */

long resource_start_config(ipc_server_config_type *config, Resource *resource)
{
  resource->activate();
  return ipc_server_start_config(config);
}



/* A finaliser for exposed resources. */

void resource_thread_finaliser(ipc_server_config_type *config)
{
  Resource *resource = reinterpret_cast<Resource *>(config->finaliser_obj);

  /* Close and delete the resource. */

  resource->close();
  delete resource;

  /* Release the capabilities. */

  ipc_server_finalise_config(config);
  delete config;
}
